/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */

package osmq.agents;

import java.io.*;
import java.net.*;
import java.util.*;

import osmq.util.*;
import osmq.mque.MQue;
import osmq.logs.Logs;
import osmq.net.SocketStreamFilter;

/**
  * Base class for socket server class that listens for publisher and subscriber
  * connections, passing them to a dedicated thread.
  */
class ReaderConnectionManager extends Threaded
   {
   private boolean           _pausechildren = false;
   private boolean           _shutdownall = false;
   private ExceptionListener _el = null;

   private ServerSocket      _serversocket = null;
   private Socket            _workersocket = null;

   private static int        _instance = 0;
   private int               _portID = 0;
   private boolean           _binarymode;
   private MQue              _pipeline;
   private Vector            _handlers = new Vector();
   private Accumulator      _accum = new Accumulator();

   ReaderConnectionManager(MQue queue)
       {
       _pipeline = queue;
       }

   Vector getHandlers(){return _handlers;}
   boolean isShutdownAll(){return _shutdownall;}
   Accumulator getAccumulator(){return _accum;}

   /**
     * if <i>binary</i> is <i>true</i>, the socket will
     * be opened in binary content mode.  Otherwise, socket
     * will be opened in text mode.
     */
   public void setBinarySocketMode(boolean binary)
      {_binarymode = binary;}
   public boolean isBinarySocketMode()
      {return _binarymode;}

   public void setPortID(int v)
     {_portID = v;}
   public int getPortID()
     {return _portID;}

   public void setExceptionListener(ExceptionListener listener)
      {
      _el = listener;
      }

   public ExceptionListener getExceptionListener()
      {
      return _el;
      }

   protected void initializer() throws Exception
      {
      if(getPortID() < 1)
         throw new IOException("Invalid Port ID");
      if(_serversocket == null)
         _serversocket = new ServerSocket(getPortID());
      }

   protected void iterativeProcess() throws Exception
     {
     // Accept a new socket connection
     try{_workersocket = _serversocket.accept();}

     // If there was a socket exception, close the
     // newly created socket
     catch(SocketException v)
        {
        try{_workersocket.close();}
        catch(Exception e){} // close the worker socket
        Logs.err.println("Error in Server.accept() within ConnectionManager: "
                         + v.getMessage());
        return;
        }

      catch(Exception v)
           {
           Logs.err.println("Error in ConnectionManager: " +
                              v.getMessage());
           return;
           }

     try{
        launchWorker();
        }
     catch(Exception e)
        {
        try{_workersocket.close();}
        catch(Exception ignore){}
        if(_el != null)
           _el.onException(e);
        // implicit return without throwing exception
        }
     }


   private void launchWorker() throws Exception
      {
      SocketStreamFilter bs = new SocketStreamFilter(_workersocket, isBinarySocketMode());

      // create an instance of that worker
      ReaderConnectionHandler worker = new ReaderConnectionHandler(_pipeline, this);

      // set the worker's socket stream filter
      worker.setStreamFilter(bs);
      worker.start();
      }

   public void setPauseHandlers(boolean pausechildren)
      {
      _pausechildren = pausechildren;
      }

   public boolean isPauseHandlers()
      {return _pausechildren;}

   public int getConnectionCount()
     {
     return getHandlers().size();
     }

   protected void cleanup()
      {
      try
        {
        // close the listening socket
        if(_serversocket != null)
           _serversocket.close();
        }
      catch(Exception e) {}
      finally {_serversocket = null;}
      }

  public long getMessageCount()
       {return _accum.getCount();}

  public  void killAllPublishers()
      {
      _shutdownall = true;
      killGroup();
      }

    protected void killGroup()
      {
      try
         {
         Iterator i = _handlers.iterator();
         if(i != null)
           while(i.hasNext())
             {
             ReaderConnectionHandler sh = (ReaderConnectionHandler) i.next();
             if(sh != null)
               {
               try{sh.stop();}
               catch(Exception e){}
               }
            }
         i = null;
         }
      catch(Exception e)
         {
         Logs.err.println("Error in killGroup " + e.getMessage());
         }
      }

}


