/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */


package osmq.io;

import java.net.*;
import java.io.*;
import java.util.*;
import java.math.*;
import osmq.util.SpecialChars;

public class StringInputStream extends FilterInputStream implements ReaderStream
   {
   final private static int SIZE_INCREMENT = 16;
   private int arraysize = SIZE_INCREMENT; // intially set to 16, and grows by increments of 16;
   private int off;
   private int readbyte;
   private byte[] _buffer = new byte[SIZE_INCREMENT];
   private String rv;

   public StringInputStream(InputStream in)
      {super(in);}

   private void increaseBufferSize()
      {
      arraysize += SIZE_INCREMENT;
      byte[] newbuffer = new byte[arraysize];
      System.arraycopy(_buffer, 0, newbuffer, 0, arraysize - SIZE_INCREMENT);
      _buffer = newbuffer;
      }

    public int readBytes(byte[] target) throws IOException
       {
       off = 0;              // initialize index

       while(true)
          {
          // and read the next byte from the stream
          readbyte = in.read();

          // while the return byte does not indicate EOF and
          // is not a newline, append it to the array
          switch(readbyte)
            {
            case -1:
               throw new IOException("End of input stream");

            case SpecialChars.STRING_TERMINATOR_BYTE:  // string terminator
               if(off == 0)
                {
                // first element is a newline, so ignore
                readbyte = in.read();
                continue;
                }
             else
                {
                return off;
                }

             default:
                // increase the array size if the array would overflow
                if(off == arraysize)
                    throw new IOException("target array too small");

                // add the byte to the array ...
                target[off++] = (byte) readbyte;
             }
          }
       }


    public byte[] readBytes() throws IOException
       {
       off = 0;              // initialize index

       while(true)
          {
          // and read the next byte from the stream
          readbyte = in.read();

          // while the return byte does not indicate EOF and
          // is not a newline, append it to the array
          switch(readbyte)
            {
            case -1:
               throw new IOException("End of input stream");

            case SpecialChars.STRING_TERMINATOR_BYTE:  // string terminator
               if(off == 0)
                {
                // first element is a newline, so ignore
                readbyte = in.read();
                continue;
                }
             else
                {
                byte[] returnpacket = new byte[off];
                System.arraycopy(_buffer, 0, returnpacket, 0, off);
                return returnpacket;
                }

             default:
                // increase the array size if the array would overflow
                if(off == arraysize)
                     increaseBufferSize();

                // add the byte to the array ...
                _buffer[off++] = (byte) readbyte;
             }
          }
       }



    public String readln() throws IOException
       {
       off = 0;              // initialize index

       while(true)
          {
          // and read the next byte from the stream
          readbyte = in.read();

          // while the return byte does not indicate EOF and
          // is not a newline, append it to the array
          switch(readbyte)
            {
            case -1:
               throw new IOException("End of stream");

            case SpecialChars.STRING_TERMINATOR_BYTE:  // string terminator
               if(off == 0)
                {
                // first element is a newline, so ignore
                readbyte = in.read();
                continue;
                }
             else
                return new String(_buffer, 0, off);

             default:
                // increase the array size if the array would overflow
                if(off == arraysize)
                     increaseBufferSize();

                // add the byte to the array ...
                _buffer[off++] = (byte) readbyte;
             }
          }
       }

   /**
     * Reads <code>len</code> bytes of data from the input stream into
     * an array of bytes.
     * The number of bytes actually read is returned as an integer.
     * <p>
     * <i>NOTE</i>  This method blocks until input <code>len</code> is available,
     * end of file is detected, or an exception is thrown.  This is different from the normal InputStream contract that
     * may return less than <code>len</code>
     *
     * <p> If <code>b</code> is <code>null</code>, a
     * <code>NullPointerException</code> is thrown.
     *
     * <p> If <code>off</code> is negative, or <code>len</code> is negative, or
     * <code>off+len</code> is greater than the length of the array
     * <code>b</code>, then an <code>IndexOutOfBoundsException</code> is
     * thrown.
     *
     * <p> If <code>len</code> is zero, then no bytes are read and
     * <code>0</code> is returned; otherwise, there is an attempt to read at
     * least one byte. If no byte is available because the stream is at end of
     * file, the value <code>-1</code> is returned; otherwise, at least one
     * byte is read and stored into <code>b</code>.
     *
     * <p> The first byte read is stored into element <code>b[off]</code>, the
     * next one into <code>b[off+1]</code>, and so on. The number of bytes read
     * is, at most, equal to <code>len</code>. Let <i>k</i> be the number of
     * bytes actually read; these bytes will be stored in elements
     * <code>b[off]</code> through <code>b[off+</code><i>k</i><code>-1]</code>,
     * leaving elements <code>b[off+</code><i>k</i><code>]</code> through
     * <code>b[off+len-1]</code> unaffected.
     *
     * <p> In every case, elements <code>b[0]</code> through
     * <code>b[off]</code> and elements <code>b[off+len]</code> through
     * <code>b[b.length-1]</code> are unaffected.
     *
     * <p> If the first byte cannot be read for any reason other than end of
     * file, then an <code>IOException</code> is thrown. In particular, an
     * <code>IOException</code> is thrown if the input stream has been closed.
     *
     * <p> The <code>read(b,</code> <code>off,</code> <code>len)</code> method
     * for class <code>InputStream</code> simply calls the method
     * <code>read()</code> repeatedly. If the first such call results in an
     * <code>IOException</code>, that exception is returned from the call to
     * the <code>read(b,</code> <code>off,</code> <code>len)</code> method.  If
     * any subsequent call to <code>read()</code> results in a
     * <code>IOException</code>, the exception is caught and treated as if it
     * were end of file; the bytes read up to that point are stored into
     * <code>b</code> and the number of bytes read before the exception
     * occurred is returned.  Subclasses are encouraged to provide a more
     * efficient implementation of this method.
     *
     * @param      b     the buffer into which the data is read.
     * @param      off   the start offset in array <code>b</code>
     *                   at which the data is written.
     * @param      len   the maximum number of bytes to read.
     * @return     the total number of bytes read into the buffer, or
     *             <code>-1</code> if there is no more data because the end of
     *             the stream has been reached.
     * @exception  IOException  if an I/O error occurs.
     * @see        java.io.InputStream#read(byte[], int, int)
     */
   public int read(byte[] b, int off, int len) throws IOException
       {
       int n = 0;
       while (n < len)
          {
          // read into the array bytes from the stream, for a length
          // not to exceed the balance of len (not already read)
          int count = this.in.read(b, off + n, len - n);
          if (count < 0)
             {
             return -1;
             }
          n += count;
          }
      return len;
      }
   }
