/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */

package osmq.mcns;

import java.net.*;
import java.io.*;
import osmq.util.Threaded;

/**
  * A threaded base class for respondants to requests on a multicast channel.
  * The class answers Muticast requests on a specific channel and passes the
  * string to an onRequest() function that returns a string.  If the String
  * returned by onRequest() is not null, it is returned to the requestor node
  * via a datagram write.
  */

public abstract class MCResponder  extends Threaded
  {
  private int _channel = MCNS.DISCOVERY_CHANNEL;
  private String  _dottedgroupaddress;
  private int     _ttl;
  private String  _returnstring;

  // the MulticastListener contains a multicast socket
  private MulticastListener _listener = null;

  private boolean _isactive = false;

  // The response socket is not a multicast socket.  Rather, it is
  // a standard datagram socket that is used to send an acknowledgement
  // message back to the the requestor.
  private DatagramSocket _responsesocket;
  private DatagramPacket _packet;
  private byte           _returnbuf[];

  protected MCResponder()
     {
     _dottedgroupaddress = NetworkScope.getScope().getGroupAddress();
     _ttl = NetworkScope.getScope().getTtl();
     _channel = getDefaultChannel();
     }

  abstract protected int getDefaultChannel();

  /** passed the incoming datagram message.
    * what is returned by this function is returned to the node
    * that sent the message
    */
  abstract protected String onRequest(String service);

  public static byte buf[] = new byte[MCNS.READBUF_SIZE];

  public boolean isActive(){return _isactive;}

  protected void initializer() throws IOException
    {
    this.setPriority(Threaded.VERY_HIGH_PRIORITY);
    _listener = new MulticastListener();
    _listener.setChannel(this.getChannel());
    _listener.setDottedGroupAddress(this.getDottedGroupAddress());
    _listener.setTimeToLive(this.getTimeToLive());
    _responsesocket = new DatagramSocket();
    _isactive = true;
    }

  public void setChannel(int channel)
     {_channel = channel;}

  public int getChannel()
     {return _channel;}

  public void setDottedGroupAddress(String dottedaddress)
     {_dottedgroupaddress = dottedaddress;}

  public String getDottedGroupAddress()
     {return _dottedgroupaddress;}


  public void setTimeToLive(int ttl)
     {_ttl = ttl;}

  public int getTimeToLive()
     {return _ttl;}


  final protected void iterativeProcess()  throws IOException
     {
     try{
        _listener.getNextPacket();

        if(_listener.isTimedOut() || _listener.isEmptyPacket())
           return;

        // inherited class determines response, if any
        _returnstring = onRequest(_listener.getRemoteData());

        // Only respond to request if onRequest did not return null
        if(_returnstring == null)
          return;

        // make a byte array from the message buffer
        _returnbuf = _returnstring.getBytes();

        _packet = new DatagramPacket(_returnbuf,
                                    _returnbuf.length,
                                    _listener.getRemoteAddress(),
                                    _listener.getRemotePort());
        _responsesocket.send(_packet);
        }

     catch(Exception e)
          {
          System.err.println("exiting main thread in MCResponder on an error: " +
                             e.getMessage());
          System.out.println("Exiting MultiChannel Responder " +
                                           e.getMessage());
          throw new MCNSException(e.getMessage());
          }
     }


  protected synchronized void cleanup()
     {
     try{
        if(_listener != null)
           _listener.close();
        }
     catch(Exception e){}
     try{
        if(_responsesocket != null)
           _responsesocket.close();
        }
     catch(Exception e){}
     finally
        {
        _listener = null;
        _responsesocket = null;
        }
     }
  }
