/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */

package osmq.messages;

import java.util.*;
import java.io.Serializable;

import osmq.records.*;
import osmq.util.Immutable;
import osmq.util.ImmutableException;
import osmq.util.ImmutableImpl;
import osmq.util.ByteArrayImpl;
import osmq.util.ByteArray;
import osmq.util.DynByteArray;

/**
  * A BinaryMessageBean is an implementation of the Message
  * interface, and an alternative to the ERMessageBean implementation
  * (DATASET).  It has header attributes, but unlike the ERMessage,
  * the body is not considered a record with multiple elements,
  * but rather contains an array of bytes.
  */
public class BinaryMessageBean extends ImmutableImpl implements
                                  Message, Cloneable, Serializable
  {
  private ByteArray     _body = null;
  private MessageHeader _header = null;
  private DynByteArray  _fullmessage;

  public BinaryMessageBean()
    {
    this(new MessageHeader(), (ByteArray)null);
    }

   BinaryMessageBean(MessageHeader header, ByteArray body)
     {
     _body = body;
     _header = header;
     }

   BinaryMessageBean(MessageHeader header, byte[] body)
     {
     _body = new ByteArrayImpl(body);
     _header = header;
     }

   BinaryMessageBean(String bodytext)
     {
     _body = new ByteArrayImpl(bodytext.getBytes());
     createAHeader();
     }

   BinaryMessageBean(byte[] body)
     {
     _body = new ByteArrayImpl(body);
     createAHeader();
     }

   public void setImmutable()
     {
     super.setImmutable();
     if(_header != null)
        _header.setImmutable();
     }

   public void setHeader(MessageHeader header)
     {
     errorIfImmuatable();
     _header = header;
     }

   public MessageHeader getHeader()
     {
     return _header;
     }

    public byte[] getHeaderBytes()
      {
      return _header.getBytes();
      }

   public void setHeaderBytes(byte[] value)
      {
      errorIfImmuatable();
      _header = new MessageHeader(value);
      }

   public byte[] getClientBytes()
      {
      return _body.getBytes();
      }

   public void setClientBytes(byte[] data)
      {
      errorIfImmuatable();
      _body = new ByteArrayImpl(data);
      }

   public void setClientData(ByteArray ds)
      {
      _body = ds;
      }

   public String getMessageAsString()
      {
      return new String(getMessageAsBytes());
      }

   /**
     * Returns the topic of the message.
     *
     * @return  A string containing the topic. The topic defines the
     * message contents for a subscriber (number of elements and the
     * data type of each element).
     */
   public String getTopic()
      {
      return _header.getTopic();
      }

   public void setTopic(String topic)
      {
      _header.setTopic(topic);
      }

   /**
     * Returns the unique identifier for the data contained in the message.
     * (This "key" is relative to the the topic.)
     *
     * @return  A string containing the unique identifier.
     */
   public String getKey()
      {
      return _header.getKey();
      }

   public void setKey(String s)
      {
      _header.setKey(s);
      }

   public void setKeyAsInt(int n)
      {
      _header.setKeyAsInt(n);
      }

   public int getKeyAsInt()
      {
      return _header.getKeyAsInt();
      }

   /**
     * Returns the identifier of the event or activity that caused the
     * transaction to be generated.  This is a value that is based on the topic.
     * <p>
     * Standard events used for down-stream subscribers that maintain
     * databases are ERMessage.EVENT_ADD, ERMessage.EVENT_DELETE
     * and ERMessage.and EVENT_UPDATE.
     *
     * @return  A string containing the transaction event.
     */
   public String getTransaction()
     {return _header.getTransaction();}

   /**
    * Returns the first character of the event identifer, or a null character if
    * there is no event identifer
    */
    public char getTransactionAsChar()
       {
       return _header.getTransactionAsChar();
       }


    public void setTransaction(String e)
       {
       _header.setTransaction(e);
       }

    public void setTransactionAsChar(char c)
       {
       _header.setTransactionAsChar(c);
       }

    /**
      * Gets the identifier used to identify a group of transactions
      * as related.  The relationship is likely a unit-of-work that
      * resulted in tranaction-related topics being generated.
      * This is an optional header attribute.
      *
      * @return  An int containing the correlation identifier.
      */
    public long getCorrelationNumber()
       {
       return _header.getCorrelationNumber();
       }

    public void setCorrelationNumber(long n)
       {
       _header.setCorrelationNumber(n);
       }

    public String getCorrelationString()
       {
       return _header.getCorrelationString();
       }

    public void setCorrelationString(String s)
       {
       _header.setCorrelationString(s);
       }

    /**
      * Returns the identifier of the message sender (publisher).  The
      * sender identifies a unique client process (instance).
      *
      * @return  A string containing the message sender.
      */
    public String getOriginator()
       {
       return  _header.getOriginator();
       }

    public void setOriginator(String t)
       {
       _header.setOriginator(t);
       }

    /**
      * Returns the identifier of the message recipient, or
      * null if the message is intended for all subscriber to the
      * message topic.
      *
      * @return  A string containing the recipient identifier.
      */
    public String getRecipient()
       {
       return _header.getRecipient();
       }

    /**
      * Sets the unique recipient that will receive the message.
      * If a recipient value is set, only that recipient will receive
      * the message. (Point-to-point messaging).  Setting the recipient
      * should be reserved for clients requesting a service (type of
      * TYPE_REQUEST) and for OLTP systems that receive a request that
      * is being rejected.  In the latter case, the OLTP system would send
      * a point-to-point rejection message to the originator of the request.
      *
      * @param value the intended message recipient
      */
    public void setRecipient(String t)
      {
      _header.setRecipient(t);
      }

    /**
      * Identifies whether the message type is a NOTIFICATION or REQUEST .
      *
      * @return  A char containing either the value TYPE_NOTIFICATION or
      * TYPE_REQUEST.
      */
    public char getType()
      {
      return _header.getType();
      }

    /**
      * Sets the event type. The message type is the category
      * of event that caused a message to be generated.
      * Standard types include  MessageAttributes.TYPE_NOTIFICATION
      * and MessageAttributes.TYPE_REQUEST.
      * <p>TYPE_NOTIFICATION messages are published by OLTP systems
      * after it commits a transaction. (Batch applications can also
      * publisher TYPE_NOTIFICATION messages.)
      * <p>TYPE_REQUEST messages are <i>point-to-point</i> routed client
      * requests for service.  (<i>Point-to-point</i> messages are sent to a
      * single recipient, and have the recipient identified in the message
      * header, verus <i>publish-subscribe</i> messages that have no recipeint and
      * are published to all subscribers to the message topic.)
      * @see #setRecipient(String t)
      *
      */
    public void setType(char t)
       {
       _header.setType(t);
       }

   /**
     * Sets the time the message was publisherd.  This
     * function is called by the message publisher after
     * client submits the message for delivery.
     * <p><B>NOTE</B> This value should be set by the underlying
     * client middleware, and not the application.
     *
     * @param n the time as a long.
     */
   public void setTimeStamp(long n)
     {
     _header.setTimeStamp(n);
     }

   /**
     * Sets the sequence number for this message.
     * <p><B>NOTE</B> This value should be set by the underlying
     * client middleware, and not the application.
     *
     * @param value the new sequence number
     */
  public void setSequenceNumber(long n)
     {
     _header.setSequenceNumber(n);
     }

  /**
    * Returns a timestamp indicating when the message was sent.
    *
    * @return  A long containing the time in milli-seconds.
    */
  public long getTimeStamp()
     {
     return _header.getTimeStamp();
     }

  /**
    * Returns the relative sequence number of this message.
    * When a message is sent by a client, it denotes the relative
    * sending sequence.  When a message is received, it denotes
    * the relative sequence of all messages fetched by the recipient,
    * regardless of the sender.
    *
    * @return  A long identifying the message sequence.
    */
   public  long getSequenceNumber()
      {
      return _header.getSequenceNumber();
      }

    /**
      *  Identifies the body format as either text or binary.
      */
    public char getFormat()
       {
       return _header.getFormat();
       }


    public void setFormat(char t)
       {
       _header.setFormat(t);
       }

   /**
     * Determine if there is a recipient identifier value.
     *
     *
     * @return  true if there is a non-null recipient identifier.
     */
  public boolean hasAssignedRecipient()
      {
      return _header.hasAssignedRecipient();
      }

   /**
     * Determines if there is a topic identifier value.
     *
     * @return  true if there is a non-null topic identifier.
     */
   public boolean hasAssignedTopic()
      {
      return _header.hasAssignedTopic();
      }

   /**
     * Determines if there is a correlation identifier value.
     *
     * @return  true if there is a non-null numeric correlation identifier.
     */
   public boolean hasAssignedCorrelationNumber()
      {
      return _header.hasAssignedCorrelationNumber();
      }

   /**
     * Determines if there is a key value identifier.
     *
     * @return  true if there is a non-null key identifier.
     */
   public boolean hasAssignedKey()
      {
      return _header.hasAssignedKey();
      }

    /**
    * Determines whether a position has a value (versus null)
    *
    * @param p the position containing the value to be tested
    * @return  true of there is a non-null value at position p, else false
    * @exception  ArrayIndexOutOfBoundsException if p is not a valid element position.
    */

   public boolean hasAssignedOriginator()
      {
      return _header.hasAssignedOriginator();
      }

   public byte[] _duplicate(byte[] original)
      {
      int count = original.length;
      byte result[] = new byte[count];
      System.arraycopy(original, 0, result, 0, count);
      return result;
      }

  private void createAHeader()
     {
     _header = new MessageHeader();
     }

  public byte[] getMessageAsBytes()
     {
     int addedbytesize = (_body.hasADelimiter()) ? 0 : 1;
     String hd = _header.getText();
     int hdsize = hd.length();

     // If body doesn't have a delimiter already, a byte must
     // be reserved for it
     byte[] result  = getFullMessageArray(_body.getBytes().length +
                       hdsize + addedbytesize).getBuffer();

     System.arraycopy(hd.getBytes(), 0, result, 0, hdsize);

     // If the body does not alreay provide a delimiter,
     // append the header's delimiter character in the position
     // immediately following the header array
     if(_body.hasADelimiter())
        result[hdsize] = (byte) hd.charAt(0);

     // The offset into the target where the body will be appended
     // depends on whether there was a delimiter byte already appended
     System.arraycopy(_body.getBytes(),
                      0,
                      result, hdsize + addedbytesize,
                      _body.getBytes().length);
     return result;
     }

  public Object clone()
        {
        BinaryMessageBean bpm = (BinaryMessageBean) super.clone();
        bpm._header = (MessageHeader) _header.clone();
        bpm._body = (_body == null || _body.getBytes().length == 0) ?
              null : new ByteArrayImpl(_duplicate(_body.getBytes()));
        return bpm;
        }

   private void errorIfImmuatable()
      {
       if(isImmutable())
         throw new ImmutableException("Message is immutable");
      }

   private DynByteArray getFullMessageArray(int sz)
       {
       if(_fullmessage == null)
           _fullmessage = new DynByteArray(sz);
       else
          _fullmessage.initializeBuffer(sz);
       return _fullmessage;
       }

   public static void main(String[] args)
      {
      DataRecord mds = new DataRecordBean("|hello|how|are|you");
      int max = 10000;
      Vector messages = new Vector();
      BinaryMessageBean dcm = new BinaryMessageBean();
      dcm.setRecipient("RECIP_BOB");
      dcm.setTopic("TOPIC_CUSTOMER");
      for(int ctr = 0; ctr < max; ++ctr)
         {
         String svalue = new String("" + ctr);

         //dcm.setClientData(svalue.getBytes());
         dcm.setClientData(mds);
         dcm.setSequenceNumber(ctr);
         messages.addElement(dcm.clone());
         }

      while(messages.size() > 0)
         {
         BinaryMessageBean m = (BinaryMessageBean) messages.elementAt(0);
      //   System.out.println("Sequence is " + m.getAttributes().getSequenceNumber());
      //   System.out.println("Value is " + new String(m.getClientData()));
         System.out.println("Message string is " + m.getMessageAsString());
         messages.removeElementAt(0);
         }
      System.out.println("The End");
      }
   }














