/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */


package osmq.messages;

import java.math.BigDecimal;
import java.util.Date;
import osmq.records.*;
import osmq.util.KeyedData;
import osmq.util.Immutable;
import osmq.util.ImmutableException;

/**
  * A multi-element data set (record) that includes identifying attributes
  * such as a topic, action, data key, and correlation ID.
  * Application data elements are referenced positionally, using an API
  * similar to that empolyed by JDBC. An element can be put or retrieved as
  * a String, int, double, long or char.
  * The first element is reference as position 1, and the last as position n.
  */

public interface MessageAttributes extends KeyedData, Immutable
   {
   final public static int TOTAL_ATTRIBUTES = 10;
   // ----------------------------------------------------------------------------------
   // FUTURE RELEASE WILL ADD A BROKER(SERVER) IDENTIFIER.  THIS IDENTIFIES THE
   // SERVER THAT FIRST RECEIVED THE CLIENT MESSAGE.  MESSAGES ARE ROUTED FROM THE
   // SERVER TO REMOTE SERVERS, BUT ONLY MESSAGES THAT ORIGINATED AT THE CURRENT SERVER
   // ----------------------------------------------------------------------------------
   // ? WILL THE SERVER PREPEND THE HEADER ATTRIBUTE, OR WILL THE ATTRIBUTE
   // BE BLANKK (TO BE FILLED IN)
   // ?? WILL THE SIZE OF THE ENTRY BE PREALLOCATED (SO THE BYTE ARRAY DOESN'T NEED TO
   // CHANGE IN SIZE)

   // HEADER POSITIONAL ELEMENTS
   // Note that the front-most positions are those that are parsed by the broker.
   final public static int RECIPIENT       = 1;  // Unique process id that maps into a host/port address
   final public static int TOPIC           = 2;  // Ideally the name of a database table
   final public static int TYPE            = 3;  // N = Notification of transaction event
                                                 // R = Request for service
                                                 // D = Denial of service request (response)
                                                 // T = Topic change request
                                                 // B = Batch Control message
   final public static int ORIGINATOR      = 4;  // Unique process ID
   final public static int CONTENT_FORMAT  = 5;  // U = unknown,
                                                 // T = text,
                                                 // B = binary,
                                                 // D = Dataset
   final public static int DATA_KEY        = 6;  // Optional (for loading keyed arrays)
   final public static int TRANSACTION     = 7;  // A = add
                                                 // U = update
                                                 // D = delete
   final public static int CORRELATION_ID  = 8;  // 0-n Unit of work group identifier.
                                                 // If present, subsequent messages with
                                                 // the same ID are considered part of the same
                                                 // unit of work.
   final public static int SEND_TIMESTAMP  = 9;  // 19980409162306
                                                 // CCYYMMDDHHMMSS
   final public static int SENDER_SEQ      = 10; // Generated by publisher adapter (0-n)

   // TRANSACTION EVENT TYPICAL VALUES
   final public static char TRANS_ADD          = 'A';
   final public static char TRANS_DELETE       = 'D';
   final public static char TRANS_UPDATE       = 'U';


   /**
     * Returns the topic of the message.
     *
     * @return  A string containing the topic. The topic defines the
     * message contents for a subscriber (number of elements and the
     * data type of each element).
     */
   String getTopic();

   void setTopic(String topic);

   /**
     * Returns the unique identifier for the data contained in the message.
     * (This "key" is relative to the the topic.)
     *
     * @return  A string containing the unique identifier.
     */
   String getKey();
   void setKey(String s);

   void setKeyAsInt(int n);
   int  getKeyAsInt();

   /**
     * Returns the identifier of the event or activity that caused the
     * transaction to be generated.  This is a value that is based on the topic.
     * <p>
     * Standard events used for down-stream subscribers that maintain
     * databases are TRANS_ADD, TRANS_DELETE, and TRANS_UPDATE.
     *
     * @return  A string containing the transaction event.
     */
   String getTransaction();

   /**
    * Returns the first character of the transaction identifer, or a null character if
    * there is no transaction identifer.
    *
    * @return  A char containing the transaction event.
    */
    char getTransactionAsChar();

    void setTransaction(String e);
    void setTransactionAsChar(char c);

    /**
      * Gets the message attribute used to identify a group of related
      * transactions.  The relationship is likely a unit-of-work that
      * resulted in tranaction-related topics being generated.
      * This is an optional header attribute.
      *
      * @return  An int containing the correlation identifier.
      */
    long getCorrelationNumber();
    void setCorrelationNumber(long n);

    String getCorrelationString();
    void setCorrelationString(String s);

    /**
      * Returns the identifier of the message sender (publisher).  The
      * sender identifies a unique client process (instance).
      *
      * @return  A string containing the message sender.
      */
    String getOriginator();


    /**
      * Sets the identifier of the message sender/originator.
      * Setting the originator should be performed when a message is published,
      * and this value defaults to the client identifier used to attach to the
      * message server.
      * <p><b>NOTE</b> This method should not be called by an application, since
      * the value will be overridden by the message publisher adapter.
      *
      * @param value the sender/originator of the message
      */
    void setOriginator(String t);

    /**
      * Returns the identifier of the message recipient, or
      * null if the message is intended for all subscriber to the
      * message topic.
      *
      * @return  A string containing the recipient identifier.
      */
    String getRecipient();

    /**
      * Sets the unique recipient that will receive the message.
      * If a recipient value is set, only that recipient will receive
      * the message. (Point-to-point messaging).  Setting the recipient
      * should be reserved for clients requesting a service (event type of
      * MessageType.REQUEST) and for OLTP systems that receive a request that
      * is being rejected.  In the latter case, the OLTP system would send
      * a point-to-point rejection message to the originator of the request.
      *
      * @param value the intended message recipient
      */
    void setRecipient(String t);

    /**
      * Identifies whether the message type is a MessageType.NOTIFICATION
      * or MessageType.REQUEST .
      *
      * @return  A char containing the event type value.
      * @see #setType(char)
      */
    char getType();

    /**
      * Sets the message type. The message type is the category
      * of event that caused a message to be generated.
      * Standard type include MessageType.NOTIFICATION ('N')
      * and MessageType.REQUEST ('R'), MessageType.DENIAL ('D'),
      * MessageType.TOPIC_CHG ('T') and MessageType.BATCH_CONTROL ('B').
      * <p>TYPE_NOTIFICATION messages are published by OLTP systems
      * after it commits a transaction. (Batch applications can also
      * publisher MessageType.NOTIFICATION messages.)
      * <p>TYPE_REQUEST messages are <i>point-to-point</i> routed client
      * requests for service.  (<i>Point-to-point</i> messages are sent to a
      * single recipient, and have the recipient identified in the message
      * header, verus <i>publish-subscribe</i> messages that have no recipeint and
      * are published to all subscribers to the message topic.)
      * <p> MessageType.DENIAL messages are a point-to-point messages, routed to the originator
      * by the service provider.  They indicate a request for service has been denied.
      * <p> MessageType.TOPIC_CHG messages are used internally to identify a TOPIC CHANGE EVENT.
      * <p> MessageType.BATCH_CONTROL message are sent immediately before and after a batch of
      * related messages.  They are used to delimit the batch, so that subscribers
      * can determine an action to be taken (such as creating export files, or updating database
      * tables).  The terminating CONTROL message will normally contain totals
      * that should be considered a checksum.
      * @see #setRecipient(String t)
      *
      */
    void setType(char t);

   /**
     * Sets the time the message was publisherd.  This
     * function is called by the message publisher after
     * client submits the message for delivery.
     * <p><B>NOTE</B> This value should be set by the underlying
     * client middleware, and not the application.
     *
     * @param n the time as a long.
     */
   void setTimeStamp(long n);

   /**
     * Sets the sequence number for this message.
     * <p><B>NOTE</B> This value should be set by the underlying
     * client middleware, and not the application.
     *
     * @param value the new sequence number
     */
  void setSequenceNumber(long n);

  /**
    * Returns a timestamp indicating when the message was sent.
    *
    * @return  A long containing the time in milli-seconds.
    */
    long getTimeStamp();

  /**
    * Returns the relative sequence number of this message.
    * When a message is sent by a client, it denotes the relative
    * sending sequence.  When a message is received, it denotes
    * the relative sequence of all messages fetched by the recipient,
    * regardless of the sender.
    *
    * @return  A long identifying the message sequence.
    */
    long getSequenceNumber();

    /**
      *  Identifies the body format as either dataset, text or binary.
      *  Valid values include
      * <p> FORMAT_UNKNOWN ('U')
      * <p> FORMAT_TEXT    ('T')
      * <p> FORMAT_DATASET ('D')
      * <p> FORMAT_BINARY  ('B')
      */
    char getFormat();


    /**
      * Sets the attributes that identifies the type of data contained
      * in the message body.
      * <p> Valid values include
      * <p>   FORMAT_UNKNOWN ('U')
      * <p>   FORMAT_TEXT    ('T')
      * <p>   FORMAT_DATASET ('D')
      * <p>   FORMAT_BINARY  ('B')
      * <p> ERMessages default to a format of FORMAT_DATASET.
      */
    void setFormat(char t);

   // --------------------- TEST IF ATTRIBUTE IS SET ----------------------

   /**
     * Determine if there is a recipient identifier value.
     *
     *
     * @return  true if there is a non-null recipient identifier.
     */
   boolean hasAssignedRecipient();

   /**
     * Determines if there is a topic identifier value.
     *
     * @return  true if there is a non-null topic identifier.
     */
   public boolean hasAssignedTopic();

   /**
     * Determines if there is a correlation identifier value.
     *
     * @return  true if there is a non-null numeric correlation identifier.
     */
   public boolean hasAssignedCorrelationNumber();

   /**
     * Determines if there is a key value identifier.
     *
     * @return  true if there is a non-null key identifier.
     */
   public boolean hasAssignedKey();

    /**
    * Determines whether a position has a value (versus null)
    *
    * @param p the position containing the value to be tested
    * @return  true of there is a non-null value at position p, else false
    * @exception  ArrayIndexOutOfBoundsException if p is not a valid element position.
    */

   boolean hasAssignedOriginator();

   Object clone();
   }



