/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */

package osmq.mque;

import java.util.*;
import java.io.*;
import osmq.util.*;

  /**
    * A thread-safe, byte-oriented FIFO (first-in first-out) queue that
    * holds byte array and/or String elements.
    * An instance can have one or more <i>writing</i> threads, but limited to
    * only one <i>reader</i>.
    * <p>This class is an alternative to an OS <i>pipe</i>, since a Queue can hold
    * any number of objects, limted only by its implementation.
    * <p> The implementation  includes the following aggreagate classes:
    * <p> 1) a <b>MemoryFifo</b> (memory-resident) and
    * <p> 2) a <b>DiskFifo</b> (file-resident)
    * <p> The <b>MQue</b> uses the <i>MemoryFifo</i> exclusively
    *        until a user-definable threshold capacity is reached.  It then utilizes
    *        the <i>DiskFifo</i> to hold the excess entries.
    * <P> A <b>MemoryFifo</b> is instantiated in an opened state. with a
    * <b>DiskFifo</b> being created only if/when necessary.
    * @see MemoryFifo
    * @see DiskFifo
    * <p><p>

    * <b>Basic Algorithms</b>
    * Read priority is
    * <p> 1) putback element
    * <p> 2) disk
    * <p> 3) memory
    * <p>Always write to the <b>MemoryFifo</b>, with a test before the write to
    * determine whether memory is full.  If it is full, the memory elements
    * are appended to the <b>DiskFifo</b>.
    */

public class MQue implements Writable, EventCounter, Serializable
  {
  private static int DEFAULT_PAGING_THRESHOLD = 1000;
  private static int REBUILD_FACTOR = 50;
  private long        _totalbytes = 0;
  private byte[]      _pulledbytes;
  private MemoryFifo  _memory = new MemoryFifo();   // ram fifo
  private DiskFifo    _disk = null;                 // disk fifo -- instantiated as needed
                                                    // when ram size exceeds threshold, items
                                                    // are written to disk

                  // ram fifo size threshold
  private int     _pagingthreshold = DEFAULT_PAGING_THRESHOLD;

  private int     _maxthreshold = 0;

  private boolean _isopen = true;
  private boolean _writehalted = false;
  private long    _count = 0;     // current element count
  private long    _highcount = 0; // highest number of elements on the queue at one time
  private long    _pushcount = 0; // total number of elements written to the queue
  private long    _pullcount = 0; // total number of elements read from the queue
  private byte[]  _pushbackbytes = null;
  private String  _instancename = null;

  public MQue()
     {
     this("");
     }

  public MQue(String name, long startcount)
     {
     this(name);
     // The queue is empty, but appears to have had
     // messages already written and read
     _pushcount = startcount;
     _pullcount = startcount;
     }

  public MQue(String name)
     {
     _instancename = name;  // optional instance name
     }

  /**
    * Returns a name associated with this MQue instance.
    */
   public String getInstanceName()
     {
     return _instancename;
     }

  /**
    * Sets a name associated with this MQue instance.
    */
   public void setInstanceName(String instancename)
     {
     _instancename = instancename;
     }

   public String toString()
     {
     return getInstanceName();
     }
     
  /**
    * Returns the cumulative length (in bytes) of all elements
    * that have been written to the queue.
    */
  final public long getBytesWritten()
      {
      return _totalbytes;
      }

  /**
    * Writes a String to the queue.
    * @throws IOException if queue is in a <i>closed</i> state,
    *         if the queue is in a <i>write-halt</i> state, or
    *         if the total number of elements in the queue would
    *         exceed the max allowable count.
    */
  public void writeString(String s) throws IOException
      {
      // No need to duplicate bytes since String is immutable
      _putBytes(s.getBytes());
      }
      
  /**
    * Creates a duplicate copy of the byte array and writes it
    * to the queue.
    * @throws IOException if queue is in a <i>closed</i> state,
    *         if the queue is in a <i>write-halt</i> state, or
    *         if the total number of elements in the queue would
    *         exceed the max allowable count.
    */
  public void writeBytes(byte[] s) throws IOException
     {
     _putBytes(createDuplicateArray(s));
     }

  /**
    * Writes the byte array to the queue without duplicating it.
    * Although this is a more rapid method than writeBytes(byte[]),
    * it may cause unpredicable results if the array is later changed.
    * @throws IOException if queue is in a <i>closed</i> state,
    *         if the queue is in a <i>write-halt</i> state, or
    *         if the total number of elements in the queue would
    *         exceed the max allowable count.
    */
  public void writeBytesNoDuplicate(byte[] s) throws IOException
     {
     _putBytes(s);
     }

  /**
    * Returns the next element in the queue as a byte array.  This is
    * a non-blocking read of the queue: if the queue is empty,
    * this method returns a value of <i>null</i>.
    * @throws IOException if queue is in a <i>closed</i> state and the
    *         queue is <i>empty</i>.
    */
  public synchronized byte[] readBytesNoWait() throws IOException
      {
      if(isClosed() && _count == 0)
         throw new IOException("Attempt to read from closed empty queue");

      if(_pushbackbytes != null)
         {
         _pulledbytes = _pushbackbytes;
         _pushbackbytes = null;
         }
      else if(_disk != null)
         {
         _pulledbytes = _disk.readBytes();
         destroyEmptyDisk();
         }
      else if(!_memory.isEmpty())
         {
         _pulledbytes = _memory.readBytes();
         }
      else
         _pulledbytes = null;

      if(_pulledbytes != null)
         {
         _count--;
         _pullcount++;
         }
      return _pulledbytes;
      }

  /**
    * Returns the next element in the queue as a byte array.  If the
    * queue is empty, the method blocks until one is available or
    * the queue is closed.
    * @throws IOException if queue is in a <i>closed</i> state and the
    *         queue is <i>empty</i>.
    */
  public synchronized byte[] readBytes() throws IOException
      {
      // If queue is empty, normally go into a wait state
      while(_count == 0)
         {
         // If queue is empty and the queue is closed,
         // an exception is thrown
         ensureOpen();

         // this is open and no elements, so wait
         try{wait();}
         
         catch(InterruptedException e) // alert after notify
            {
            continue; // so loop back and test again
            }
         }
      // There are elements, so read them
      // (even if the queue is closed)
      return readBytesNoWait();
      }
      
  /**
    * Empty records from the queue until the pulled count
    * equals a specific value
    */
  public synchronized boolean flushToPullCount(long count) throws IOException
     {
     while(!isEmpty() && _pullcount < count)
        readBytes();
     return isEmpty();
     }

  /**
    * Clears the queue of all elements.
    */
  public synchronized void clear() throws IOException
     {
     while(!isEmpty())
         readBytes();
     }

  /**
    * Empties the queue, but maintains the pushed/pulled statistics.
    */
  protected synchronized void fastClear() throws IOException
     {
     _pullcount += _count;  // clear the remainder, so they will be considered pulled
     _memory.clear();       // empty the memory queue
     _disk.close();         // discard the disk object
     _disk = null;
     _pushbackbytes = null; // discard any pushback value
     _count = 0;            // reset the count to zero
     }

  /**
    * Returns <i>true</i> if the queue is in an opened state.
    * Otherwise, returns <i>false</i>.
    */
  public synchronized boolean isOpen()
     {
     return _isopen;
     }

  /**
    * Returns <i>true</i> if the queue has been closed.
    * Otherwise, returns <i>false</i>.
    */
  public boolean isClosed()
     {
     return !isOpen();
     }

  /**
    * Reopens a queue that has been closed.
    */
  public synchronized void open()
     {
     if(_isopen)
         return;
     _count = 0;
     _pushbackbytes = null;
     _isopen = true;
     _writehalted = false;
     }

  /** Halts all writes, resulting in a QueueException being thrown on any
    * subsequent calls to the Queue object's write methods.
    * When MQue's size drops to zero, <i>read</i> method calls will
    * throw a QueueException.
    */
  public synchronized void close()
     {
     _isopen = false;
     _writehalted = true;
     if(_count == 0)
        notify(); // notify reader blocked on a read
     }


  /**
    * Pulls elements from another queue, writing them to this queue.
    */
  public synchronized void writeAllFromQue(MQue f) throws IOException
     {
     while(!f.isEmpty())
         writeBytes(f.readBytes());
     }

  /**
    * Returns the next element on the queue as a String.  This is
    * a non-blocking read of the queue: if the queue is empty,
    * this method returns a value of <i>null</i>.
    * @throws IOException if queue is in a <i>closed</i> state and the
    *         queue is <i>empty</i>.
    */
  public String readStringNoWait() throws IOException
      {
      byte[] b = readBytesNoWait();
      return (b == null) ? null : new String(b);
      }

  /**
    * Returns the next element on the queue as a String.  If the
    * queue is empty, the method blocks until an element is available or
    * the queue is closed.
    * @throws IOException if queue is in a <i>closed</i> state and the
    *         queue is <i>empty</i>.
    */
  public String readString() throws IOException
      {
      return new String(readBytes());
      }

  /**
    * Returns an element that was pulled from the queue.  This element
    * will be the next to be read from the queue.
    * @throws IOException if no read has been performed since the
    * last occurance of a putBack(), or if the queue is closed.
    */
  public synchronized void putBack(byte[] o) throws IOException
      {
      ensureOpen();
      if(_pushbackbytes != null)
         throw new QueueException("Attempting to put back twice without a pull",
                                   QueueException.QE_DOUBLE_PUSH_BACK);
      _pushbackbytes = o;
      _count++;
      _pullcount--;
      }

  /**
    * Returns <i>true</i> if the queue size (element count) is zero.
    */
  public boolean isEmpty()
     {return _count == 0;}

  /**
    * Returns the next element from the queue.  This is an alias
    * method for readBytes().
    * @see #readBytes()
    */
  public byte[] nextElement()
      {
      try{
         return readBytes();
         }
      catch(IOException e)
         {
         return null;
         }
      }

  /**
    * Returns the number of elements currently on the queue.
    */
  public long size()
     {
     return _count;
     }

  /**
    * Returns the highest queue count (elements on the queue)
    * since the queue was instantiated or since the max count
    * was cleared.
    * @see #clearHighestSize()
    */
  public long highestSize()
     {return _highcount;}

  /**
    * Clears the highest count.  A call to highSize() immediately
    * after calling this method returns a value of zero.
    * @see #highestSize()
    */
  public void clearHighestSize()
     {_highcount = 0;}

  /**
    * Returns the number of elements that have been written to the
    * queue since it was instantiated.
    */
  public long getPushCount()
     {return _pushcount;}
     
  /**
    * Returns the number of elements that have been read from the
    * queue since it was instantiated.
    */
  public long getPullCount()
     {
     return _pullcount;
     }

  /**
    * Returns <i>true</i> if the queue is in a <i>write halt</i> state.
    * Otherwise, returns <i>false</i>.
    */
  public boolean isHaltWrites()
     {
     return _writehalted;
     }

  /**
    * Sets the write halt mode.  When set to <i>true</i>, any attempt to
    * write an element to the queue will return an exception.
    */
  public synchronized void setHaltWrites(boolean halt)
     {
	   _writehalted = halt;
	   }

  /**
    * Returns <i>true</i> if the queue is in a read halt state.
    * A read halt state is when the queue is closed and the queue
    * is empty. Attempting to read an element in a <i>read halt</i> state
    * throws an exception.
    */
  public boolean isHaltReads()
     {
     return !_isopen &&
             _count == 0;
     }

  /**
    * Sets the element count threshold, where overflow elements are
    * paged to disk.
    * @see #getPagingThreshold()
    */
  public synchronized void setPagingThreshold(int v)
     {
     _pagingthreshold = v;
     }

  /**
    * Returns the paging threshold (in elements).
    * @see #setPagingThreshold(int)
    */
  public int getPagingThreshold()
     {
     return _pagingthreshold;
     }

  /**
    * Sets the maximum capacity threshold.  When the queue
    * reaches this capacity (element count), subsequent writes
    * throw an exception.
    * A value of 0 indicates no threshold.
    */
  public synchronized void setMaxThreshold(int threshold)
     {
     _maxthreshold = (threshold > -1) ? threshold : _maxthreshold;
     }

   /**
    * Returns the maximum capacity threshold (element count).
    * When the queue reaches this capacity, subsequent writes
    * throw an exception.
    * A value of <i>0</i> indicates no threshold.
    */
  public int getMaxThreshold()
     {
     return _maxthreshold;
     }
     
  /**
    * Returns the number of element read events.  This is an
    * alias for getPullCount().  This method supports EventCounter
    * interface.
    */
  public long getEventCount()
     {
     return getPullCount();
     }

  private synchronized void _putBytes(byte[] s) throws IOException
      {
      // throw an exception if closed or in a write halt state
      validateWriteState();

      if(_maxthreshold > 0 &&
         _count > _maxthreshold)
         {
         throw new IOException("Attempt to exceed MQue max threshold");
         }

      // if paging threshold exceed, flush memory queue to disk
      pageOnThreshold();
      _memory.writeBytes(s);

      _count++;
      _pushcount++;
      _totalbytes += s.length;

      if(_count > _highcount)
         _highcount = _count;

      if(_count == 1)
        notify();    // notify any reader

      return;
      }

  private boolean isReadWaitState()
     {
     return isEmpty() && isOpen();
     }

  private void destroyEmptyDisk()
      {
      if(_disk != null && _disk.size() == 0)
         {
         _disk.close();
         _disk = null;
         }
      }

  private byte[] createDuplicateArray(byte[] b)
     {
     if(b == null || b.length < 1)
         throw new IllegalArgumentException("Null byte array");
     byte[] rv = new byte[b.length];
     System.arraycopy(b, 0, rv, 0, b.length);
     return rv;
     }
     
  private void ensureOpen() throws IOException
      {
      if(!_isopen)
         throw new QueueException("Queue is closed", QueueException.QE_CLOSED);
      }

  // queue must be open and not in a write halt state
  private void validateWriteState() throws IOException
      {
      ensureOpen();
      errorOnWriteHalt();;
      }

  private void errorOnWriteHalt() throws IOException
      {
      if(_writehalted)
	      throw new QueueException("Queue in write halt mode", QueueException.QE_WRITE_HALTED);
      }
      
  private void pageOnThreshold()
      {
      // If the memory fifo size exceeds the designated threshold,
      // first append to the disk fifo all the elements that are in the memory fifo
      // (emptying the memory fifo), then begin writing elements to the disk fifo.
      if(_memory.size() > _pagingthreshold)
         {
         try{
            // If there is no disk fifo, build one
            if(_disk == null)
               _disk = new DiskFifo();
               
            // Append all memory messages to disk fifo,
            // resulting in an empty memory fifo
            _disk.appendFrom(_memory);

            // Periodic rebuild of a fractured disk fifo
            rebuildFracturedDiskFifo();
            }
         catch(IOException e){}
         }
      }

  private void rebuildFracturedDiskFifo()
     {
     try{
        // If the number of objects pulled from the disk fifo exceeds
        // 50 times the memory threshold, the fifo is considered "fractured" and
        // needs rebuilding
        // (to free up disk space occupied by the pulled objects"
        if(_disk.getPulledCount() > (getPagingThreshold() * REBUILD_FACTOR))
            {
            System.out.println("Rebuilding");
            DiskFifo tempdisk = new DiskFifo();
            tempdisk.appendFrom(_disk);
            _disk.close();
            _disk = tempdisk;
            }
         }
     catch(IOException e){}
     }

  }
