/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */

package osmq.net;

import java.net.*;
import java.io.*;
import osmq.util.*;
import osmq.mcns.*;

/**
  * An aggregate class of ServerSocket.class.  MCNSServer is a seperately threaded
  * class that answers name service requests from client applications.
  * MCNSServer listens for discovery inquiry datagrams on a published
  * address /port and responds if the request is for the service provided by the
  * ServerSocket instance.
  */

public class DiscoveryServer extends MCResponder
  {
  private StateEventBean seb = new StateEventBean();
  private String _myname = null;
  private int _myport = 0;
  String  _mymode;

  private String _messageString; // concatenated message

  private DiscoveryServer()
     {
     super();
     }

  protected int getDefaultChannel()
     {return MCNS.DISCOVERY_CHANNEL;}
     
  public void start(String servicename, int portaddress, boolean binarymode) throws IOException
     {
     // Store service name and test that no other server is running
     // with that name.
     buildServiceMessage(servicename, portaddress, binarymode);

     try
        {
        // start listening for discovery inquiries
        start();
        }
     catch(Exception e)
       {
       // an unexptected listening failure, so stop the thread
       stop();

       // notify any listener about the failure
       seb.fireFailEvent();
       }
     }

  public static DiscoveryServer createAServer(ExceptionListener listener)
     {
     DiscoveryServer ds = new DiscoveryServer();
     if(listener != null)
        ds.setExceptionListener(listener);

     return ds;
     }

  public static DiscoveryServer createAndStart(Object caller,
                                               String servername,
                                               int portaddress,
                                               int _ttl,
                                               boolean binarymode)
                                               throws IOException
     {
     DiscoveryServer ds = new DiscoveryServer();

     if(_ttl > -1)
        ds.setTimeToLive(_ttl);
     if(caller != null)
        {
        if(caller instanceof ExceptionListener)
           ds.setExceptionListener((ExceptionListener) caller);

        if(caller instanceof StateEventListener)
           ds.addStateEventListener((StateEventListener) caller);

        }

     ds.start(servername, portaddress, binarymode);
     return ds;
     }

  public synchronized void addStateEventListener(StateEventListener l)
     {
     seb.addStateEventListener(l);
     }

  public synchronized void removeStateEventListener(StateEventListener l)
     {
     seb.removeStateEventListener(l);
     }

   public int getPortID()
     {
     return _myport;
     }

  public String getServiceID()
     {
     return _myname;
     }

  /**
    * If passed string is my service name , return a string containing
    * my service name  and IP address.  Nomatch return null,
    * resulting in nothing being returned to the requestor process.
    */
  public String onRequest(String passed)
    {
    if(passed.toUpperCase().equals(_myname))
        {
        return _messageString;
        }
    else
        return null;
     }

  private void buildServiceMessage(String servicename, int portaddress, boolean binarymode) throws IOException
     {
     try
        {
        _myname = servicename;
        _myport = portaddress;
     //   _myaddress = InetAddress.getLocalHost().getHostAddress();
        _mymode =  binarymode ? "B" : "T";

        // make cetain name is a valid string
        if(servicename == null || servicename.length() < 1)
           throw new MCNSException("Invalid service name");

        // force it into upper case
        servicename = servicename.toUpperCase();

        // Make certain no current instance of this service is running
        if(DiscoveryClient.getSingleton().isActiveService(getServiceID()))
            throw new MCNSException("Service " + getServiceID() + " already running");

        // create the string that will be returned to discovery clients
        _messageString = new String(
                                    getServiceID() +
                                    ' ' +
                                    InetAddress.getLocalHost().getHostAddress() +
                                    ' ' +
                                    portaddress +
                                    ' ' +
                                    _mymode +
                                    ' ');

        // notify any status event listener that everyting went fine
        seb.fireSuccessEvent();
        }
     catch(Exception e)
        {
        // notify any status event listener of the failure
        seb.fireFailEvent();
        throw new MCNSException(e.getMessage());
        }
     }



  }
