/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */
 
package osmq.pulsar;
  
import java.net.*;
import java.io.*;
import osmq.mcns.*;
import osmq.util.Threaded;
import osmq.util.*;
import osmq.records.*;

/**
  *
  * MCNS (multicast Name Service)
  * MulticastPulsar uses an ip multicast channel to repeatedly broadcast a service name, ip adddress and port 
  * at a fixed time frequency. This feature permits the discovery of a service without having to maintain
  * that information on an NIS or DNS table, and further provides the advantage of not requiring that a 
  * service be tied to a specific server. (Useful for implementing redundant hot backup servers).
  *
  * <p>Another use of the pulsar is to identify whether a service has failed. For example, after a client
  * discovers and attaches to a service, it can listen to the multicast port for the continued broadcast
  * and take appropriate action if and when the broadcast of that service is no longer heard.
  * Similarly, in the event of a server failure, the broadcast channel can be used by a backup server
  * to broadcast a changed ip address / port combination for the service.
  *
  * <p>Reciprocal classes include
  * <p>1) DiscoveryServer -- a service class that performs the function of a name server of MulticastPulsars.
  *                       Clients pass the name of a sercie to the DiscoveryServer and are returned the
  *                       current ip address port number for that service.
  *     It logs 
  *    all active service pulsars and acts as a centralized lookup for clients, and 
  * 2) the ServicePulsarMonitor class which a client uses to monitor the that a service 
  *    remains alive during an active session. 
  * @see osmq.net.DiscoveryServer
  */

public class DynamicPulsar extends Threaded
  {


  private StringGenerator      _sg = null;
  private DataRecordGenerator  _drg = null;

  protected DatagramSocket     _socket = null;
  private String               _groupaddressString;
  boolean                      _isheartbeat = false;
  int                          _pulsarChannel = MCNS.PULSAR_CHANNEL;
  private int                  _frequencySecs = Threshold.HEARTBEAT_FREQUENCY;
  int                          _delayMillis = _frequencySecs * MCNS.SECONDS;

  public DynamicPulsar(StringGenerator sg)
     {
     _sg = sg;
     _groupaddressString = NetworkScope.getScope().getGroupAddress();
     }
     
  public DynamicPulsar(DataRecordGenerator dsg)
     {
     _drg = dsg;
     _groupaddressString = NetworkScope.getScope().getGroupAddress();
     }

  public void setDelayInSeconds(int secs)
     {
     _frequencySecs = secs;
     _delayMillis = _frequencySecs * MCNS.SECONDS;
     }
     
  public int getDelayInSeconds()
     {
     return _frequencySecs;
     }

  public void setHeartbeat(boolean v)
     {_isheartbeat = v;}
  public boolean isHeartbeat()
     {return _isheartbeat;}
     
  public void setDiscovery(boolean v)
      {_isheartbeat = !v;}
  public boolean isDiscovery()
      {return !_isheartbeat;}

  private InetAddress groupaddress;

  DatagramPacket packet;
  Thread t;
  private int priority = Thread.NORM_PRIORITY;
 
  public void setServiceChannel(int p)
     {_pulsarChannel = p;}
  public int getServiceChannel()
     {return _pulsarChannel;}
     
  public void setGroupAddress(String a)
     {_groupaddressString = a;}
  public String getGroupAddress()
     {return _groupaddressString;}

  public void setGenerator(StringGenerator sg)
     {
     _sg = sg;
     }

  public StringGenerator getGenerator()
     {
     return _sg;
     }

  protected void iterativeProcess() throws Exception
     {
     byte[] bytes;
     if(_sg != null)
        bytes = _sg.getString().getBytes();
     else
        bytes = _drg.getDataRecord().getText().getBytes();
     packet = new DatagramPacket(bytes,
                                 bytes.length,
                                 groupaddress,
                                 _pulsarChannel);

     _socket.send(packet);
     Thread.sleep(_delayMillis);
     }

  protected void initializer() throws Exception
     {
     _socket = new DatagramSocket();
     groupaddress = InetAddress.getByName(_groupaddressString);


     System.out.println(DateStamp.getTimeStamp() +
                        " Multicasting DataRecord to " +
                        _groupaddressString +
                        " on port " +
                        _pulsarChannel +
                        " every " +
                        _frequencySecs +
                        " secs.");
     }

  protected void cleanup()
    {
    if(_socket != null)
       _socket.close();
    _socket = null;
    System.out.println(DateStamp.getTimeStamp() +
                       " Terminating DataRecord multicaat to " +
                       _groupaddressString +
                       " on port " +
                       _pulsarChannel);
    }


} 
